/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2022 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxlibs_internal_H
#define fesslix_flxlibs_internal_H

#include "flxlibs.h"


#ifdef __unix__
  #include <dlfcn.h>                // dlopen functionality
#elif defined(__WINDOWS__)
  #include <windows.h>
#endif
 
#ifdef __WINDOWS__
  typedef HMODULE handle_type;
#else
  typedef void* handle_type;
#endif



typedef const double (*flxfhndlsymb_t)(const double[], const int*);
typedef void (*flxfhndlsymbmtx_t)(const double[], const int*, double[], const int*, int*);
typedef const int (*flxfhndlsymbstrproc_t)(const char*, const int*);


class FlxLibFunManager {
  private:
    #ifdef __unix__
      std::map<std::string, void*> LibList;
    #elif defined(__WINDOWS__)
      std::map<std::string, HMODULE> LibList;
    #endif
  public:
    ~FlxLibFunManager();
    
    handle_type get_handle(const std::string& libpn);
    void insert_handle(const std::string& libpn, handle_type handle);
};




// -------------------------------------------------------------------------------

class FunLoadLibFun : public FunBaseFun_MtxConst {
  private:
    const std::string fname;
    flxfhndlsymb_t fhandle;
  public:
    FunLoadLibFun ( const std::string fname, flxfhndlsymb_t fhandle, std::list< FlxMtxConstFun* > *ParaList_Mtx) : FunBaseFun_MtxConst(ParaList_Mtx), fname(fname), fhandle(fhandle) {};
    const tdouble calc();
    const std::string write_v() { return fname; }
    virtual const bool dependOn_Const(const tdouble* const thenumber) { return false; }
};

class FunReadFunLoadLibFun : public FunReadFunBase_MtxConst {
  private:
    const std::string fname;
    flxfhndlsymb_t fhandle;
  public:
    FunReadFunLoadLibFun(const std::string fname, flxfhndlsymb_t fhandle) : fname(fname), fhandle(fhandle) {}
    FunBase* read ( bool errSerious ) { return new FunLoadLibFun( fname, fhandle, read_para(1,errSerious) ); }
};


// -------------------------------------------------------------------------------

class FlxObjLoadLibProc : public FlxObjBase {
  private:
    const std::string fname;
    flxfhndlsymbstrproc_t fhandle;
    FlxString* strV;
    void task();
  public:
    FlxObjLoadLibProc ( bool dolog, const std::string& fname, flxfhndlsymbstrproc_t fhandle, FlxString* strV ) : FlxObjBase(dolog), fname(fname), fhandle(fhandle), strV(strV) {}
    ~FlxObjLoadLibProc() { delete strV; }
};

class FlxObjReadLoadLibProc : public FlxObjReadBase {
  private:
    const std::string fname;
    flxfhndlsymbstrproc_t fhandle;
  public:
    FlxObjReadLoadLibProc(const std::string& fname, flxfhndlsymbstrproc_t fhandle) : fname(fname), fhandle(fhandle) {}
    FlxObjBase* read ();
};

// -------------------------------------------------------------------------------

class FlxObjLoadLibMtx : public FlxObjBase {
  private:
    const std::string fname;
    flxfhndlsymbmtx_t fhandle;
    FlxMtxConstFun* v;
    FlxMtxConstFun* w;
    void task();
  public:
    FlxObjLoadLibMtx ( bool dolog, const std::string& fname, flxfhndlsymbmtx_t fhandle, FlxMtxConstFun* v, FlxMtxConstFun* w ) : FlxObjBase(dolog), fname(fname), fhandle(fhandle), v(v), w(w) {}
    ~FlxObjLoadLibMtx() { delete v; delete w; }
};

class FlxObjReadLoadLibMtx : public FlxObjReadBase {
  private:
    const std::string fname;
    flxfhndlsymbmtx_t fhandle;
  public:
    FlxObjReadLoadLibMtx(const std::string& fname, flxfhndlsymbmtx_t fhandle) : fname(fname), fhandle(fhandle) {}
    FlxObjBase* read ();
};







#endif // fesslix_flxlibs_internal_H
