/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2022 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "fesslix.h"


void error_out(std::string errMsg);


int main(int argc, char** argv)
{

  try {
  
    // start the Fesslix-engine
      const int itmp = flxengine_init(true,argc, argv);
      if (itmp>=0) return itmp;
      
      GlobalVar.slog(3) << "Fesslix: path to executable is        '" << GlobalVar.get_exe_PathName() << "'" << std::endl;
      GlobalVar.slog(3) << "Fesslix: current working directory is '" << GlobalVar.pwd << "'" << std::endl;
      bool green = true;
 
    // evaluate the Command CMD
      #if FLX_BOOST_PO
        if (GlobalVar.cmd1st) {
          if (fesslix_CMDstr != "" && green) {
            green = (flx_parse_file(fesslix_CMDstr,false,false,true)==RETURN_SUCCESS);
          }
        }
      #endif
      
    // open and evaluate the file(s)
      for (tuint i=0;i<fesslix_fileVec.size();++i) {
        if (!green) break;
        green = (flx_parse_file(fesslix_fileVec[i],true,false,true)==RETURN_SUCCESS);
      }
    
    // evaluate the Command CMD
      #if FLX_BOOST_PO
        if (GlobalVar.cmd1st==false) {
          if (fesslix_CMDstr != "" && green) {
            green = (flx_parse_file(fesslix_CMDstr,false,false,true)==RETURN_SUCCESS);
          }
        }
      #endif
  
    // start the interactive mode
      if (fesslix_call_interact) {
        FlxObjInteractive ObjI = FlxObjInteractive(true,"cout",!FlxVi_Deactivate);
          ObjI.exec();
        GlobalVar.logLevel_strong_reset();
      }

    // return success
      flxengine_unload();
      flx_print_RETURN_SUCCESS();
      flx_delete_FLXcout();
      if (fesslix_pauseAtEnd) { flx_take_a_break(); }
      return RETURN_SUCCESS;
  }
  catch ( FlxEndE const &e) {
    FLXMSG("main_3",1);
    GlobalVar.logLevel_strong_reset();
    GlobalVar.slog(4) << "end: 'end;' was called" << std::endl;
    flxengine_unload();
    flx_print_RETURN_SUCCESS();
    flx_delete_FLXcout();
    return RETURN_SUCCESS;
  }
  catch ( FlxException &e) {
    FLXMSG("main_4",1);
    error_out(e.what());
    return RETURN_ERROR;
  }
  catch (std::exception const &e) {
    FLXMSG("main_5",1);
    error_out(std::string("ERROR: Whoops, something went wrong: ") + e.what());
    return RETURN_ERROR;
  }
  catch(...) { 
    FLXMSG("main_6",1);
    error_out("ERROR: whoops!");
    return RETURN_ERROR;
  }

}


void error_out(std::string errMsg)
{
  // write to the actual error stream
    *(GlobalVar.get_true_cerr()) << std::endl << errMsg << std::endl;
  // write to the default error stream of Fesslix
    if (GlobalVar.get_true_cerr() != GlobalVar.get_cerr() ) {
      *(GlobalVar.get_cerr()) << std::endl << errMsg << std::endl;
    }
  // write to the log-file
    GlobalVar.logLevel_strong_reset();
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << errMsg << std::endl;
      flx_print_RETURN_ERROR();
    }
  flxengine_unload();
  flx_delete_FLXcout();
  if (fesslix_pauseAtEnd) { flx_take_a_break(); }
}


